#HEADER
#                      arg/Backend/argLaTeXBackend.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

import numbers
import os
import platform
import re
import sys

import pylatex as pl
import yaml
from pylatex.utils import NoEscape, bold, italic, verbatim

from arg.Common.argMultiFontStringHelper import argMultiFontStringHelper
from arg.Backend.argBackendBase import argBackendBase

# Load supported colors
common_dir = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(common_dir, "../Common/argTypes.yml"),
          'r',
          encoding="utf-8") as t_file:
    supported_types = yaml.safe_load(t_file)

# Retrieve supported verbosity levels
colors = supported_types.get(
    "FontColors")


class plAbstract(pl.base_classes.Environment):
    """A class representing the abstract as a LaTeX environment
    """

    _latex_name = "abstract"


class argLaTeXBackend(argBackendBase):
    """A concrete class providing a LaTeX backend
    """

    # Map from greek letters to their respective unicodes
    GreekLetters = {
        "Alpha": r"A",
        "Beta": r"B",
        "Gamma": r"\Gamma",
        "Delta": r"\Delta",
        "Epsilon": r"E",
        "Zeta": r"Z",
        "Eta": r"H",
        "Theta": r"\Theta",
        "Iota": r"I",
        "Kappa": r"K",
        "Lambda": r"\Lambda",
        "Mu": r"M",
        "Nu": r"N",
        "Xi": r"\Xi",
        "Omicron": r"O",
        "Pi": r"\Pi",
        "Rho": r"P",
        "Sigma": r"\Sigma",
        "Tau": r"T",
        "Upsilon": r"\Upsilon",
        "Phi": r"\Phi",
        "Chi": r"X",
        "Psi": r"\Psi",
        "Omega": r"\Omega",
        "alpha": r"\alpha",
        "beta": r"\beta",
        "gamma": r"\gamma",
        "delta": r"\delta",
        "epsilon": r"\epsilon",
        "zeta": r"\zeta",
        "eta": r"\eta",
        "theta": r"\theta",
        "iota": r"\iota",
        "kappa": r"\kappa",
        "lambda": r"\lamdda",
        "mu": r"\mu",
        "nu": r"\nu",
        "xi": r"\xi",
        "omicron": r"o",
        "pi": r"\pi",
        "rho": r"\rho",
        "sigma": r"\sigma",
        "tau": r"\tau",
        "upsilon": r"\upsilon",
        "phi": r"\phi",
        "chi": r"\chi",
        "psi": r"\psi",
        "omega": r"\omega"}

    # List of special characters
    SpecialCharacters = "%^_#$"

    def __init__(self, parameters=None):

        # Call superclass init
        super().__init__(parameters)

        # Define backend type
        self.Type = "LaTeX"

        self.colors = None
        self.__init_colors()

    def __init_colors(self):
        # Load supported colors
        common_dir = os.path.dirname(os.path.realpath(__file__))
        with open(os.path.join(common_dir, "../Common/argTypes.yml"),
                  'r',
                  encoding="utf-8") as t_file:
            supported_types = yaml.safe_load(t_file)

        # Retrieve supported verbosity levels
        self.colors = supported_types.get("FontColors")

    def create_document_preamble(self, version=None):
        """Create document and its preamble given an Assembler version
        """

        self.create_LaTeX_output()
        self.add_packages()
        self.generate_details()
        self.add_details()
        self.generate_title_page(version)

    def create_LaTeX_output(self, document_class, document_options):
        """Create a LaTeX output file
        """

        # Prepare report prefix and output directory
        if not os.path.isdir(self.Parameters.OutputDir):
            # Try to create output directory if it does not exist yet
            print("[argLaTeXBackend] Creating output directory {}".format(self.Parameters.OutputDir))
            try:
                os.makedirs(self.Parameters.OutputDir, 0o750)
            except OSError:
                print("** ERROR: could not create output directory. Exiting.")
                sys.exit(1)

        # Create LaTeX output
        self.Report = pl.Document(
            default_filepath=os.path.join(
                self.Parameters.OutputDir,
                self.Parameters.FileName),
            documentclass=document_class,
            document_options=document_options)

        # No paragraph indentation
        self.Report.change_length(r"\parindent", "0pt")


    def add_comment(self,  comments_dict, key):
        """Add comment to the report from a dict as either text string
        or as sub-paragraph depending on number of dict values (1 or 2)
        """

        # Retrieve comment for given key and bail out if none found
        comment = comments_dict.get(str(key))
        if not comment:
            return False

        # Insert either sub-paragraph or string
        if len(comment) > 1:
            self.Report.append(
                pl.Command("subparagraph", comment[0] + ":"))
            self.Report.append(NoEscape(comment[1]))
        else:
            self.Report.append(NoEscape(comment[0]))

        # Comment was found and inserted
        return True


    def add_packages(self):
        """Add LaTeX packages
        """

        # Use richer mathematical environment
        self.Report.preamble.append(pl.Command("usepackage", "amsthm,amsmath,amssymb,stmaryrd"))

        # Make better looking tables
        self.Report.preamble.append(pl.Command("usepackage", "booktabs"))

        # Allow for multi-rows in tables
        self.Report.preamble.append(pl.Command("usepackage", "multirow"))

        # Make better looking captions
        self.Report.preamble.append(pl.Command("usepackage", "caption", NoEscape(r"width=.75\textwidth")))

        # Expand range of acceptable figure names
        self.Report.preamble.append(pl.Command("usepackage", "grffile"))

        # Let verbatim text overflow
        self.Report.preamble.append(pl.Command("usepackage", "spverbatim"))

        # Improve handling of floats
        self.Report.preamble.append(pl.Command("usepackage", "float"))

        # Insert spaces correctly after macros
        self.Report.preamble.append(pl.Command("usepackage", "xspace"))

        # Allow for slanted looking fractions
        self.Report.preamble.append(pl.Command("usepackage", "xfrac"))

        # Fix problems with underscores
        self.Report.preamble.append(pl.Command("usepackage", "underscore", "strings"))

        # Handle path with url package
        self.Report.preamble.append(pl.Command("usepackage", "url", ["obeyspaces", "spaces"]))

    def generate_details(self):
        """Generate details: title, organization, address, authors, numbers, etc.
        """

        pass

    def add_details(self):
        """Add details: title, number, issue, versions, authors, organizations, location, date, markings, etc.
        """

        # Title
        self.Report.preamble.append(pl.Command(
            "title",
            bold(NoEscape(self.Parameters.Title))))

        # Authors
        if not self.Parameters.AllAuthors:
            self.Report.preamble.append(pl.Command(
                "author",
                NoEscape(", ".join(self.Parameters.Authors))))
        else:
            self.Report.preamble.append(pl.Command(
                "author",
                NoEscape(r"{}\\{}".format(self.Parameters.AllAuthors, self.Parameters.Address))))
            # Date
        self.Report.preamble.append(pl.Command(
            "date",
            NoEscape(self.Report.Date)))

        # Include classification markings
        if self.Parameters.MarkingsFile:
            self.Report.preamble.append(pl.Command("input", NoEscape(self.Parameters.MarkingsFile)))

    def generate_title_page(self, version=None):
        """Generate title page
        """

        # Generate title pages
        self.Report.append(NoEscape(r"\maketitle"))

        # Add document timestamp
        self.add_document_provenance(version)
        # Now begin with main document body
        if self.Parameters.Abstract:
            self.Report.append(pl.Command("clearpage"))
            with self.Report.create(plAbstract()) as abstract:
                with open(os.path.join(self.Parameters.DataDir,
                                       self.Parameters.Abstract),
                          'r') as f:
                    for row in f:
                        abstract.append(NoEscape(row))

        # Start new page and append tables of contents, figures, tables
        self.add_document_tocs()

        # Start new page for acknowledgments when present
        if self.Parameters.Thanks:
            self.Report.append(pl.Command("clearpage"))
            self.add_section({"title": "Acknowledgments",
                              "include": self.Parameters.Thanks},
                             False)

        # Append preface if defined
        if self.Parameters.Preface:
            self.Report.append(pl.Command("clearpage"))
            self.add_section({"title": "Preface",
                              "include": self.Parameters.Preface},
                             False)

        # Append executive summary if defined
        if self.Parameters.ExecutiveSummary:
            self.Report.append(pl.Command("clearpage"))
            self.add_section({"title": "Executive Summary",
                              "include": self.Parameters.ExecutiveSummary},
                             False)

        # Append nomenclature if defined
        if self.Parameters.Nomenclature:
            self.Report.append(pl.Command("clearpage"))
            self.add_section({"title": "Nomenclature",
                              "include": self.Parameters.Nomenclature},
                             False)

        # Start next report core with clear double pages
        self.Report.append(pl.Command("cleardoublepage"))

    def substitute_variable_values(self, src_file, dst_file, marker, ignore_comments=True):
        """Substitute variables with corresponding values given marker
        """

        # Build regular expression for sought variables
        regexp = re.compile(marker + r"(.*?)" + marker)

        # Iterate over source file lines
        with open(src_file, 'r') as f_in:
            with open(dst_file, 'w') as f_out:
                # Iterate over all input lines and substitute as needed
                for l in f_in:
                    # Ignore commented out lines when requested
                    if ignore_comments and re.match('%', l.strip()):
                        continue

                    # Perform variable substitutions
                    for attr in regexp.findall(l):
                        # Check whether substitution can be performed
                        if hasattr(self.Parameters, attr):
                            rep_key = marker + attr + marker
                            rep_val = getattr(self.Parameters, attr)
                            if rep_val:
                                # Perform value substitution
                                print("[argLaTeXBackend]Base Substituting {} with {}".format(
                                    rep_key,
                                    rep_val))
                                l = l.replace(rep_key, rep_val)
                            else:
                                # Perform empty substitution
                                print("*  WARNING: empty value for {} attribute.".format(attr))
                                l = l.replace(rep_key, '')

                        else:
                            # Attribute does not exist in parameters
                            print("*  WARNING: unrecognized variable @@{}@@ cannot be substituted.".format(attr))

                    # Print resulting line into destination file
                    f_out.write(l)

    @staticmethod
    def generate_text(text, type="default"):
        """Add page break to the report
        """

        # Append clearpage command in LaTeX
        return "\\texttt{{{}}}".format(text)

    def generate_multi_font_string(self, multi_font_string, _=None):
        """Generate LaTeX markup from multi-font string helper instance
        """

        # Produce a string decorated with LaTeX markups
        decorated_string = r''

        # Iterate over string map
        for string, font_bits, color in multi_font_string.iterator():
            # Distinguish between supported bits
            if font_bits & 16 == 16:
                # Map Greek letter to unicode when requested
                string = "${}$".format(
                    self.GreekLetters.get(string, ''))

            else:
                # For all other cases escape LaTeX special characters
                for c in self.SpecialCharacters:
                    if c in string:
                        string = string.replace(c, "\\" + c)

                if font_bits & 1 == 1:
                    string = italic(string)
                if font_bits & 2 == 2:
                    string = bold(string)
                if font_bits & 4 == 4:
                    string = r"\texttt{{{}}}".format(string)
                if font_bits & 8 == 8:
                    string = r"$\mathcal{{{}}}$".format(string)
                if color and self.colors.get(color):
                    sep = ','
                    rgb = self.colors.get(color).split(sep)
                    string = r"\color[rgb]{{{}, {}, {}}}{}".format(int(rgb[0]) / 255.,
                                                                   int(rgb[1]) / 255.,
                                                                   int(rgb[2]) / 255.,
                                                                   string)

            # Append current decorated string to global one
            decorated_string += string

        # Return decorated string
        return decorated_string

    def generate_matrix_string(self, matrix):
        """Generate LaTeX markup from matrix entries
        """

        # Bail out early if vector is empty or does not have correct type
        if not matrix or not isinstance(matrix, (list, tuple)):
            return "[]"

        # Verify that matrix is well-formed
        n_cols = len(matrix[0])
        if [x for x in matrix if len(x) != n_cols]:
            print("*  WARNING: {} is not a well-formed matrix.".format(matrix))

        # Otherwrise a LaTeX array one column per dimension
        n_rows = len(matrix)
        m_string = r"$\{0}[\begin{1}{{{2}{3}{2}}}{4}\end{1}\{5}]$".format(
            "left" if n_rows > 1 else "big",
            "{array}",
            "@{}",
            'r' * n_cols,
            r"\\".join(['&'.join([r"\mathtt{{{}}}".format(x) for x in v]) for v in matrix]),
            "right" if n_rows > 1 else "big")

        # Return un-escaped decorated string
        return NoEscape(m_string)


    def add_list(self, item, number_items=False):
        """Add itemization or enumeration to the report
        """

        # Decide on whether to itemize or enumerate
        method = pl.Enumerate() if number_items else pl.Itemize()

        # Append all list items
        with self.Report.create(method) as enum_it:
            for items in item.get("items"):
                enum_it.add_item(NoEscape(items.get("string")))


    def add_paragraph(self, item):
        """Add paragraph to the report
        """

        # Insert new paragraph
        self.Report.append(pl.Command("par"))

        # Check whether a string or a file is to be included
        if "string" in item:
            # Retrieve provided string
            string = item["string"]

            # Decorate multi-font strings with LaTeX markup
            if isinstance(string, argMultiFontStringHelper):
                string = self.generate_multi_font_string(string)

            # Insert string into the report
            self.Report.append(NoEscape(string))

        # Insert verbatim fragment
        elif "verbatim" in item:
            self.Report.append(pl.Command("path", NoEscape(item["verbatim"])))

        # Insert LaTeX fragment as-is
        elif "latex" in item:
            self.Report.append(NoEscape(item["latex"]))

        # Insert contents of a file
        elif "include" in item and len(item["include"]) > 0:
            curr_item = item["include"][0]
            if "file" in curr_item:
                file_name = os.path.join(self.Parameters.DataDir, curr_item["file"])
                font = curr_item["font"]
                sep = curr_item["sep"] if "sep" in curr_item else ''
            else:
                file_name = os.path.join(self.Parameters.DataDir, item["include"])
                font = ''
            with open(file_name, 'r') as tex_file:
                # Add font command if provided
                if font:
                    if sep and len(font.split(sep)) > 0:
                        self.Report.append(NoEscape(font.split(sep)[0]))
                    else:
                        self.Report.append(NoEscape(font))
                # Read each row and add it to LaTeX output
                for line in tex_file:
                    self.Report.append(NoEscape(line))
                # Add font command if provided
                if font:
                    if sep and len(font.split(sep)) > 1:
                        self.Report.append(NoEscape(font.split(sep)[1]))

    def add_subdivision(self, item, command):
        """Add specified subdivision to the report
        """

        # Retrieve title string
        title_string = item.get("title", '')

        # Decorate multi-font strings with LaTeX markup
        if isinstance(title_string, argMultiFontStringHelper):
            title_string = title_string.execute_backend()

        # Add page break if compact paragraph to come
        if item.get("sections"):
            for i in item.get("sections"):
                page_break_commands = ["show_all_boundary_conditions", "show_all_materials", "show_CAD_metadata"]
                if i.get("name") in page_break_commands:
                    self.add_page_break()

        # Create subdivision and insert header if provided
        self.Report.append(pl.Command(command, NoEscape(title_string)))
        self.add_paragraph(item)

    def add_subtitle(self, item):
        """Add subsection to the report
        """

        # Call appropriate subdivision method
        self.add_subdivision(item, '*')

    def add_subsection(self, item, numbered=True):
        """Add subsection to the report
        """

        # Call appropriate subdivision method
        self.add_subdivision(item, "subsection" + ('' if numbered else '*'))

    def add_section(self, item, numbered=True):
        """Add section to the report
        """

        # Call appropriate subdivision method
        self.add_subdivision(item, "section" + ('' if numbered else '*'))

    def add_chapter(self, item, numbered=True):
        """Add chapter to the report
        """

        # Call appropriate subdivision method
        self.add_subdivision(item, "chapter" + ('' if numbered else '*'))

    def add_page_break(self):
        """Add page break to the report
        """

        # Append clearpage command in LaTeX
        self.Report.append(pl.Command("clearpage"))

    def add_table(self, header_list, body, caption_string, do_verbatim=False, pos="ht!"):
        """Add table to the report
        """

        # Bail out early if header is empty
        try:
            n_cols = len(header_list)
            if not n_cols:
                return
        except:
            return

        # Bail out early if unsupported body type
        if not body:
            return
        if type(body) == dict:
            is_dict = True
        elif type(body) == list:
            is_dict = False
        else:
            return

        # Compute row size depending on number of characters
        l_head = sum([len(h) for h in header_list])
        if is_dict:
            # Handle dict case
            l_body = max([len("{}".format(k))
                          + sum([len("{}".format(x)) for x in v if x])
                          for k, v in body.items()])
        else:
            # Handle list case
            l_body = max([sum([len(x) for x in l]) for l in body])
        l_max = n_cols + max([l_head, l_body])
        if l_max > 50:
            if l_max > 80:
                if l_max > 110:
                    row_size = NoEscape(r"\tiny ")
                else:
                    row_size = NoEscape(r"\scriptsize ")
            else:
                row_size = NoEscape(r"\footnotesize ")
        else:
            row_size = NoEscape(r"\normalsize ")

        # Create function to generate decorated strings when needed
        decorator = (lambda x: NoEscape(x.execute_backend())
        if isinstance(x, argMultiFontStringHelper)
        else x)

        # Create table
        tab_format = "@{}l" + (n_cols - 1) * 'r' + "@{}"
        with self.Report.create(pl.LongTable(table_spec=tab_format, pos=pos)) as table:
            # Create table header
            table.append(NoEscape(r"\toprule"))
            header_list = [decorator(x) for x in header_list]

            # Resize header font except for non-verbatim lists
            if is_dict or do_verbatim:
                header_list = [NoEscape("{}{}".format(row_size, h))
                               for h in header_list]
            table.add_row(list(header_list))
            table.append(NoEscape(r"\midrule\endfirsthead"))

            # Iterate over of contents to create table body
            if do_verbatim:
                # Handle contents verbatim
                if is_dict:
                    # Iterate over dictionary entries
                    for k, v in sorted(body.items()):
                        # Format value depending on it type
                        if isinstance(v, list):
                            v_f = [row_size + verbatim(
                                "{}".format(x) if x else '',
                                delimiter='@') for x in v]
                        elif isinstance(v, str):
                            v_f = [row_size + verbatim(v)]
                        elif isinstance(v, numbers.Number):
                            v_f = [row_size + "{}".format(v)]
                        else:
                            v_f = ''
                        table.add_row(
                            [row_size + verbatim("{}".format(k),
                                                 delimiter='@')]
                            + v_f)
                else:  # if not is_dict
                    # Iterate over list entries
                    for l in body:
                        table.add_row(
                            [row_size + verbatim(
                                "{}".format(x) if x else '',
                                delimiter='@') for x in l])

            else:  # if not do_verbatim
                # Handle contents depending on type
                if is_dict:
                    # Iterate over dictionary entries
                    for k, v in sorted(body.items()):
                        table.add_row(
                            [NoEscape(row_size + k)]
                            + [NoEscape("{}{}".format(row_size, x))
                               if x else '' for x in v])
                else:  # if not is_dict
                    # Iterate over list entries
                    for l in body:
                        table.add_row([decorator(x) for x in l])

            # Create table footer
            table.append(NoEscape(r"\bottomrule\\"))

            # Create table caption depending on its type
            if caption_string:
                if isinstance(caption_string, str):
                    # Directly insert base strings
                    table.append(pl.Command(
                        "caption",
                        NoEscape(caption_string)))

                elif isinstance(caption_string, argMultiFontStringHelper):
                    # Insert LaTeX string created from multi-font string
                    table.append(pl.Command(
                        "caption",
                        NoEscape(caption_string.execute_backend())))

    def add_figure(self, arguments):
        """Add figure to the report
        """

        # A figure width is required
        try:
            figure_width = arguments["width"]
        except KeyError:
            print("*  WARNING: no figure width was provided.")
            return

        # Retrieve image and associated caption for figure creation
        figure_file_name, caption_string = self.fetch_image_and_caption(
            arguments)

        # Create figure when available
        if figure_file_name:
            with self.Report.create(pl.Figure(position="ht!")) as picture:
                picture.add_image(figure_file_name, width=figure_width)

                # Append caption when available
                if caption_string:
                    if isinstance(caption_string, argMultiFontStringHelper):
                        # Convert multi-font strings to LaTeX
                        caption_string = caption_string.execute_backend()

                    # Insert caption
                    picture.add_caption(NoEscape(caption_string))

                # Insert label when available
                figure_label = arguments.get("label")
                if figure_label:
                    picture.append(pl.Command("label",
                                              NoEscape(figure_label)))

    def recursively_build_report(self, item_tree):
        """Fallback LaTeX implementation method to recursively build report
        """

        # Iterate over each dictionary in the array
        for item in item_tree:
            # Retrieve item type and determine if null marker is present
            item_type = item.get('n')
            is_numbered = not item_type.endswith("_null")

            if item_type in ("string", "latex"):
                # Directly insert text or LaTeX fragment into the report
                self.Report.append(NoEscape(item.get(item_type)))

            # Handle chapter case
            elif item_type.startswith("chapter"):
                # Create chapter
                self.add_chapter(item, is_numbered)

            # Handle section case
            elif item_type.startswith("section"):
                # Create section
                self.add_section(item, is_numbered)

            # Handle subsection case
            elif item_type.startswith("subsection"):
                # Create subsection
                self.add_subsection(item, is_numbered)

            # Handle paragraph case
            elif item_type == "paragraph":
                # Add paragraph to the report
                self.add_paragraph(item)

            # Handle itemization/enumeration case
            elif item_type in ("itemize", "enumerate"):
                # Add itemization/enumeration to the report
                self.add_list(item, True if item_type == "enumerate" else False)

            # Handle verbatim case
            elif item_type == "verbatim":
                # Include file verbatim as report paragraph
                file_name = os.path.join(self.Parameters.DataDir,
                                         item["include"])
                with open(file_name, 'r') as tex_file:
                    # Read each row and add it to LaTeX output
                    for line in tex_file:
                        self.Report.append(verbatim(NoEscape(line.strip('\n\r'))))
                        self.Report.append(NoEscape(r"\\"))

            # Handle figure case
            elif item_type == "figure":
                # Add figure to the report
                self.add_figure(item["arguments"])

            # Handle meta-information case
            elif item_type == "meta":
                # Append meta-information to the report
                self.add_meta_information(item)

            # Handle properties case
            elif item_type == "properties":
                # Append information specific to given property to the report
                self.add_information(item)

            # Handle aggregate case
            elif item_type == "aggregate":
                # Append aggregated information
                self.add_aggregation(item)

            # Proceed with recursion if needed
            if "sections" in item:
                self.recursively_build_report(item["sections"])

    def assemble(self, report_map, version=None, latex_proc=None):
        """Create LaTeX and PDF reports from given report map
        """

        # Create report with preamble
        print("[argLaTeXBackend] Generating LaTeX report")
        self.create_document_preamble(version)

        # Pass over data for report contents
        self.recursively_build_report(report_map["chapters"])

        # Append backend-specific postamble
        self.append_document_postamble()

        if self.Parameters.TexFile is not None and self.Parameters.TexFile:
            self.Report.generate_tex()
        else:
            # Generate PDF report
            print("[argLaTeXBackend] Generating PDF report")

            # Check specified LaTeX processor usability
            has_latex = self.check_latex_processor(latex_proc)
            if has_latex:
                self.Report.generate_pdf(clean_tex=False,
                                         compiler=latex_proc,
                                         compiler_args=["-pdf", "-f"],
                                         silent=True)
            # Use whatever pylatex finds if failed
            else:
                self.Report.generate_pdf(clean_tex=False,
                                         compiler_args=["-pdf", "-f"],
                                         silent=True)

    @staticmethod
    def check_latex_processor(latex_proc):
        """Check specified LaTeX processor usability
        """

        # Check specified path existence
        if latex_proc and not os.path.exists(latex_proc):
            print("*  WARNING: Specified path {} to LaTeX processor does not exist. " \
                  "Using other LaTeX processor.".format(latex_proc))
            return False

        # Check specified path leads to file
        elif latex_proc and not os.path.isfile(latex_proc):
            print("*  WARNING: Specified path {} to LaTeX processor is not a file. " \
                  "Using other LaTeX processor.".format(latex_proc))
            return False
        else:
            return True

    def add_document_provenance(self, version=None):
        """Add document provenance information on a new page
        """

        # Start new page for provenance information
        self.Report.append(pl.Command("cleardoublepage"))
        self.Report.append(NoEscape(
            r'This document was generated on {} with the Automatic Report Generator (ARG) version "{}" on the {} system {}.'.format(
                self.get_timestamp(),
                version,
                platform.system(),
                r"\texttt{{{}}}".format(platform.node()))))

    def add_document_tocs(self):
        """Add document Table of Contents, List of Figures and List of Tables on new pages
        """

        # Start new page and append tables of contents, figures, tables
        self.Report.append(pl.Command("cleardoublepage"))
        self.Report.append(pl.Command("tableofcontents"))
        self.Report.append(pl.Command("listoffigures"))
        self.Report.append(pl.Command("listoftables"))



#HEADER
#                        arg/Generation/argTools.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

def update_or_create_dict_in_dict(dict_of_dicts, key1, key2, value, update_fct):
    """Update or create dict entry in a dict with given primary and
       secondary keys, and value to be used by provided updating function
    """

    # Fetch possibly empty dict in dict of dicts
    sub_dict = dict_of_dicts.get(key1, {})

    # Invoke update function
    sub_dict[key2] = update_fct(sub_dict.get(key2), value)

    # Save current block quality stats
    return sub_dict


def map_composite_keys(in_dict, key1, separator):
    """Retrieve key1@...@keyN keys and corresponding values from input dict
       and build output dict with the form {key2:[key3, ..., keyN, value]}
       where @ denotes a 1-character separator
    """

    # Initialize top-level dict
    out_dict = {}

    # Iterate over input dict and filter by prefix with form key1@
    prefix = key1 + separator
    n = len(prefix)
    for key in [l for l in in_dict if l[:n] == prefix]:
        # Make sure that at least 2 keys were retrieved
        subkeys = key.split(separator)
        if len(subkeys) < 2:
            # Ignore non-composite keys
            continue

        # Use key2 as output dict key
        key2 = subkeys[1].strip().lower()
        out_dict[key2] = subkeys[2:] + [in_dict[key]]

    # Return result
    return out_dict



# HEADER
#                    arg/GUI/Logic/argParameterController.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
# HEADER

from PySide2.QtCore import QObject, Qt, Signal, Slot

from arg.GUI.Logic.argParameterReader import argParameterReader
from arg.GUI.Logic.argParameterWriter import argParameterWriter
from arg.GUI.Logic.argSettingsController import argSettingsController

app = "ARG-GUI"


class argParameterController(QObject):
    """A controller class to handle all available parameters
    """

    # Signals
    dataCreated = Signal(dict)
    errorDetected = Signal(str)

    def __init__(self):
        super().__init__()
        self.reader = argParameterReader(self)
        self.reader.dataCreated.connect(self.onDataCreated, Qt.DirectConnection)
        self.reader.errorDetected.connect(self.errorDetected, Qt.DirectConnection)
        self.settingsController = None
        self.writer = argParameterWriter(self)

        self.backupData = {}

    def setSettingsController(self, settingsController: argSettingsController):
        self.settingsController = settingsController

    def read(self, filePath):
        print("[{}] Reading file: '{}'".format(app, filePath))
        res = self.reader.read(filePath)
        if res and self.settingsController is not None:
            self.settingsController.setCurrentParameterFile(filePath)
        return res

    def write(self, filePath, data):
        print("[{}] Writing in file: '{}'".format(app, filePath))
        self.writer.setData(data)
        self.writer.write(filePath)
        if self.settingsController is not None:
            self.settingsController.setCurrentParameterFile(filePath)

        return True

    def reloadData(self):
        """Dictionary slot receiving created data
        """

        if self.backupData:
            self.dataCreated.emit(self.backupData)
        else:
            print("*  WARNING: No values to reload")

    @Slot(dict)
    def onDataCreated(self, data):
        """Dictionary slot receiving created data
        """

        self.backupData = data
        self.dataCreated.emit(data)



#HEADER
#                    arg/GUI/View/argGeneralOptionsWidget.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

import os

from PySide2.QtGui import QIcon
from PySide2.QtWidgets import QApplication, QFileDialog, QHBoxLayout, QCheckBox, QToolButton, QGridLayout, QLabel, \
    QLineEdit, QSizePolicy, QSpacerItem, QWidget


class argGeneralOptionsWidget(QWidget):
    """A widget class ro cover 'General Options' tab
    """

    def __init__(self):
        super().__init__()

        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        scriptDirectory = os.path.dirname(os.path.realpath(__file__))
        icon = QIcon("{}/{}".format(scriptDirectory, "../Graphics/open.png"))

        # Title
        self.titleLabel = QLabel(settings.titleValue(settings.label))
        self.titleLineEdit = QLineEdit()

        # Number
        self.numberLabel = QLabel(settings.numberValue(settings.label))
        self.numberLineEdit = QLineEdit()

        # Issue
        self.issueLabel = QLabel(settings.issueValue(settings.label))
        self.issueLineEdit = QLineEdit()

        # Version
        self.versionsLabel = QLabel(settings.versionsValue(settings.label))
        self.versionsLineEdit = QLineEdit()

        # Author
        self.authorsLabel = QLabel(settings.authorsValue(settings.label))
        self.authorsLineEdit = QLineEdit()

        # Organization
        self.organizationsLabel = QLabel(settings.organizationsValue(settings.label))
        self.organizationsLineEdit = QLineEdit()

        # Location
        self.locationLabel = QLabel(settings.locationValue(settings.label))
        self.locationLineEdit = QLineEdit()

        # Year
        self.yearLabel = QLabel(settings.yearValue(settings.label))
        self.yearLineEdit = QLineEdit()

        # Month
        self.monthLabel = QLabel(settings.monthValue(settings.label))
        self.monthLineEdit = QLineEdit()

        # Abstract File
        self.abstractFileLabel = QLabel(settings.abstractFileValue(settings.label))
        self.abstractFileLineEdit = QLineEdit()
        self.abstractFileOpenButton = QToolButton()
        self.abstractFileOpenButton.setIcon(icon)
        self.abstractFileOpenButton.clicked.connect(self.onAbstractFileOpenButtonTriggered)
        self.abstractFileHLayout = QHBoxLayout()
        self.abstractFileHLayout.addWidget(self.abstractFileLineEdit)
        self.abstractFileHLayout.addWidget(self.abstractFileOpenButton)

        # Preface
        self.prefaceLabel = QLabel(settings.prefaceValue(settings.label))
        self.prefaceLineEdit = QLineEdit()
        self.prefaceOpenButton = QToolButton()
        self.prefaceOpenButton.setIcon(icon)
        self.prefaceOpenButton.clicked.connect(self.onPrefaceOpenButtonTriggered)
        self.prefaceHLayout = QHBoxLayout()
        self.prefaceHLayout.addWidget(self.prefaceLineEdit)
        self.prefaceHLayout.addWidget(self.prefaceOpenButton)

        # Thanks
        self.thanksLabel = QLabel(settings.thanksValue(settings.label))
        self.thanksLineEdit = QLineEdit()
        self.thanksOpenButton = QToolButton()
        self.thanksOpenButton.setIcon(icon)
        self.thanksOpenButton.clicked.connect(self.onThanksOpenButtonTriggered)
        self.thanksHLayout = QHBoxLayout()
        self.thanksHLayout.addWidget(self.thanksLineEdit)
        self.thanksHLayout.addWidget(self.thanksOpenButton)

        # Executive summary
        self.executiveSummaryLabel = QLabel(settings.executiveSummaryValue(settings.label))
        self.executiveSummaryLineEdit = QLineEdit()
        self.executiveSummaryOpenButton = QToolButton()
        self.executiveSummaryOpenButton.setIcon(icon)
        self.executiveSummaryOpenButton.clicked.connect(self.onExecutiveSummaryButtonTriggered)
        self.executiveSummaryHLayout = QHBoxLayout()
        self.executiveSummaryHLayout.addWidget(self.executiveSummaryLineEdit)
        self.executiveSummaryHLayout.addWidget(self.executiveSummaryOpenButton)

        # Nomenclature
        self.nomenclatureLabel = QLabel(settings.nomenclatureValue(settings.label))
        self.nomenclatureLineEdit = QLineEdit()

        # Final
        self.finalLabel = QLabel(settings.finalValue(settings.label))
        self.finalCheckBox = QCheckBox()

        # Key Separator
        self.keySeparatorLabel = QLabel(settings.keySeparatorValue(settings.label))
        self.keySeparatorWidget = QWidget()
        self.keySeperatorLineEdit = QLineEdit()
        self.keySeperatorLineEdit.setMaximumWidth(40)
        self.keySeparatorLayout = QHBoxLayout()
        self.keySeparatorLayout.setMargin(0)
        self.keySeparatorLayout.addWidget(self.keySeperatorLineEdit)
        self.keySeparatorLayout.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Ignored))
        self.keySeparatorWidget.setLayout(self.keySeparatorLayout)

        # Instantiate 'General Options' tab layout and fill with widgets created above
        self.generalOptionsLayout = QGridLayout()
        self.generalOptionsLayout.setMargin(5)
        self.generalOptionsLayout.addWidget(self.titleLabel, 1, 0)
        self.generalOptionsLayout.addWidget(self.titleLineEdit, 1, 2)
        self.generalOptionsLayout.addWidget(self.numberLabel, 2, 0)
        self.generalOptionsLayout.addWidget(self.numberLineEdit, 2, 2)
        self.generalOptionsLayout.addWidget(self.issueLabel, 3, 0)
        self.generalOptionsLayout.addWidget(self.issueLineEdit, 3, 2)
        self.generalOptionsLayout.addWidget(self.versionsLabel, 4, 0)
        self.generalOptionsLayout.addWidget(self.versionsLineEdit, 4, 2)
        self.generalOptionsLayout.addWidget(self.authorsLabel, 5, 0)
        self.generalOptionsLayout.addWidget(self.authorsLineEdit, 5, 2)
        self.generalOptionsLayout.addWidget(self.organizationsLabel, 6, 0)
        self.generalOptionsLayout.addWidget(self.organizationsLineEdit, 6, 2)
        self.generalOptionsLayout.addWidget(self.locationLabel, 7, 0)
        self.generalOptionsLayout.addWidget(self.locationLineEdit, 7, 2)
        self.generalOptionsLayout.addWidget(self.yearLabel, 8, 0)
        self.generalOptionsLayout.addWidget(self.yearLineEdit, 8, 2)
        self.generalOptionsLayout.addWidget(self.monthLabel, 9, 0)
        self.generalOptionsLayout.addWidget(self.monthLineEdit, 9, 2)
        self.generalOptionsLayout.addWidget(self.abstractFileLabel, 10, 0)
        self.generalOptionsLayout.addLayout(self.abstractFileHLayout, 10, 2)
        self.generalOptionsLayout.addWidget(self.prefaceLabel, 11, 0)
        self.generalOptionsLayout.addLayout(self.prefaceHLayout, 11, 2)
        self.generalOptionsLayout.addWidget(self.thanksLabel, 12, 0)
        self.generalOptionsLayout.addLayout(self.thanksHLayout, 12, 2)
        self.generalOptionsLayout.addWidget(self.executiveSummaryLabel, 13, 0)
        self.generalOptionsLayout.addLayout(self.executiveSummaryHLayout, 13, 2)
        self.generalOptionsLayout.addWidget(self.nomenclatureLabel, 14, 0)
        self.generalOptionsLayout.addWidget(self.nomenclatureLineEdit, 14, 2)
        self.generalOptionsLayout.addWidget(self.finalLabel, 15, 0)
        self.generalOptionsLayout.addWidget(self.finalCheckBox, 15, 2)
        self.generalOptionsLayout.addWidget(self.keySeparatorLabel, 16, 0)
        self.generalOptionsLayout.addWidget(self.keySeparatorWidget, 16, 2)
        self.generalOptionsLayout.addItem(QSpacerItem(0, 0, QSizePolicy.Ignored, QSizePolicy.Expanding))
        self.generalOptionsLayout.setColumnMinimumWidth(1, 10)
        self.setLayout(self.generalOptionsLayout)

        # Connections
        self.abstractFileLineEdit.editingFinished.connect(self.onAbstractFileLineEditLineEditEditingFinished)
        self.prefaceLineEdit.editingFinished.connect(self.onPrefaceLineEditEditingFinished)
        self.thanksLineEdit.editingFinished.connect(self.onThanksLineEditEditingFinished)
        self.executiveSummaryLineEdit.editingFinished.connect(self.onExecutiveSummaryLineEditEditingFinished)

    def fillParameters(self, data):
        """Set all 'General Options' tab widgets to specified values
        """

        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        # Clean all widgets before filling them with appropriate values
        self.cleanParameters()

        # Title
        if data.get(settings.titleKey()):
            self.titleLineEdit.setText(data.get(settings.titleKey()))

        # Number
        if data.get(settings.numberKey()):
            self.numberLineEdit.setText(data.get(settings.numberKey()))

        # Issue
        if data.get(settings.issueKey()):
            self.issueLineEdit.setText(data.get(settings.issueKey()))

        # Version
        if data.get(settings.versionsKey()):
            self.versionsLineEdit.setText(str(data.get(settings.versionsKey())))

        # Author
        if data.get(settings.authorsKey()):
            self.authorsLineEdit.setText(data.get(settings.authorsKey()))

        # Organization
        if data.get(settings.organizationsKey()):
            self.organizationsLineEdit.setText(data.get(settings.organizationsKey()))

        # Location
        if data.get(settings.locationKey()):
            self.locationLineEdit.setText(data.get(settings.locationKey()))

        # Year
        if data.get(settings.yearKey()):
            self.yearLineEdit.setText(data.get(settings.yearKey()))

        # Month
        if data.get(settings.monthKey()):
            self.monthLineEdit.setText(data.get(settings.monthKey()))

        # Abstract
        if data.get(settings.abstractFileKey()):
            self.abstractFileLineEdit.setText(data.get(settings.abstractFileKey()))

        # Preface
        if data.get(settings.prefaceKey()):
            self.prefaceLineEdit.setText(data.get(settings.prefaceKey()))

        # Thanks
        if data.get(settings.thanksKey()):
            self.thanksLineEdit.setText(data.get(settings.thanksKey()))

        # Executive Summary
        if data.get(settings.executiveSummaryKey()):
            self.executiveSummaryLineEdit.setText(data.get(settings.executiveSummaryKey()))

        # Nomenclature
        if data.get(settings.nomenclatureKey()):
            self.nomenclatureLineEdit.setText(data.get(settings.nomenclatureKey()))

        # Final
        if data.get(settings.finalKey()):
            self.finalCheckBox.setChecked(data.get(settings.finalKey()))

        # Key Separator
        if data.get(settings.keySeparatorKey()):
            self.keySeperatorLineEdit.setText(data.get(settings.keySeparatorKey()))

    def constructParameters(self):
        """Instantiate all 'General Options' tab widgets and set to default
        """

        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        # Initialize parameters dict
        parameters = {settings.titleKey(): self.titleLineEdit.text(), settings.numberKey(): self.numberLineEdit.text(),
                      settings.issueKey(): self.issueLineEdit.text(),
                      settings.versionsKey(): self.versionsLineEdit.text(),
                      settings.authorsKey(): self.authorsLineEdit.text(),
                      settings.organizationsKey(): self.organizationsLineEdit.text(),
                      settings.locationKey(): self.locationLineEdit.text(),
                      settings.yearKey(): self.yearLineEdit.text(), settings.monthKey(): self.monthLineEdit.text(),
                      settings.abstractFileKey(): self.abstractFileLineEdit.text(),
                      settings.prefaceKey(): self.prefaceLineEdit.text(),
                      settings.thanksKey(): self.thanksLineEdit.text(),
                      settings.executiveSummaryKey(): self.executiveSummaryLineEdit.text(),
                      settings.nomenclatureKey(): self.nomenclatureLineEdit.text(),
                      settings.finalKey(): self.finalCheckBox.isChecked(),
                      settings.keySeparatorKey(): self.keySeperatorLineEdit.text()}

        return parameters

    def cleanParameters(self):
        """Reset all 'General Options' tab widgets to default
        """

        self.titleLineEdit.setText("")
        self.numberLineEdit.setText("")
        self.issueLineEdit.setText("")
        self.versionsLineEdit.setText("")
        self.authorsLineEdit.setText("")
        self.organizationsLineEdit.setText("")
        self.locationLineEdit.setText("")
        self.yearLineEdit.setText("")
        self.monthLineEdit.setText("")
        self.abstractFileLineEdit.setText("")
        self.prefaceLineEdit.setText("")
        self.thanksLineEdit.setText("")
        self.executiveSummaryLineEdit.setText("")
        self.nomenclatureLineEdit.setText("")
        self.finalCheckBox.setChecked(False)
        self.keySeperatorLineEdit.setText("")

    @staticmethod
    def onFillLineEdit(lineEdit, fileMode):
        openFileDialog = QFileDialog()
        openFileDialog.setAcceptMode(QFileDialog.AcceptOpen)
        openFileDialog.setFileMode(fileMode)
        openFileDialog.setViewMode(QFileDialog.Detail)
        if openFileDialog.exec_():
            fileNames = openFileDialog.selectedFiles()
            if len(fileNames) == 1:
                lineEdit.setText(fileNames[0])

    def onAbstractFileOpenButtonTriggered(self):
        self.onFillLineEdit(self.abstractFileLineEdit, QFileDialog.AnyFile)
        self.onAbstractFileLineEditLineEditEditingFinished()

    def onPrefaceOpenButtonTriggered(self):
        self.onFillLineEdit(self.prefaceLineEdit, QFileDialog.AnyFile)
        self.onPrefaceLineEditEditingFinished()

    def onThanksOpenButtonTriggered(self):
        self.onFillLineEdit(self.thanksLineEdit, QFileDialog.AnyFile)
        self.onThanksLineEditEditingFinished()

    def onExecutiveSummaryButtonTriggered(self):
        self.onFillLineEdit(self.executiveSummaryLineEdit, QFileDialog.AnyFile)
        self.onExecutiveSummaryLineEditEditingFinished()

    def onAbstractFileLineEditLineEditEditingFinished(self):
        QApplication.instance().checkLineEdit(self.abstractFileLineEdit, False)

    def onPrefaceLineEditEditingFinished(self):
        QApplication.instance().checkLineEdit(self.prefaceLineEdit, False)

    def onThanksLineEditEditingFinished(self):
        QApplication.instance().checkLineEdit(self.thanksLineEdit, False)

    def onExecutiveSummaryLineEditEditingFinished(self):
        QApplication.instance().checkLineEdit(self.executiveSummaryLineEdit, False)



#HEADER
#                  arg/Common/argMultiFontStringHelper.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

import os

import yaml


class argMultiFontStringHelper:
    """ A helper class to handle strings with multiple or non-default fonts
    """

    # Retrieve valid font types from YAML specification only once
    common_dir = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(common_dir, "../Common/argTypes.yml"),
              'r',
              encoding="utf-8") as t_file:
        Types = yaml.safe_load(t_file)

    print("[argMultiFontStringHelper] Supported font types: {}".format(
        ", ".join(Types.get("FontTypes", {}))))

    def __init__(self, backend=None):
        """ Constructor
        """

        # Keep track of provided backend
        self.Backend = backend

        # Create empty internal storage
        self.StringMap = []

    def __len__(self):
        """ Length
        """

        # Return sum of internal string lengths
        return sum(len(s) for (s, _, _) in self.StringMap)

    def clear(self):
        """ Clear internal storage
        """

        # Clear internal storage
        self.StringMap.clear()

    def pop(self):
        """ Pop last value from internal storage
        """

        # Safely pop value from internal storage
        return self.StringMap.pop() if self.StringMap.pop() else None

    def dump(self):
        """ Print contents of internal storage for debugging purposes
        """

        # Print internal content
        print("{} contains: {}".format(self, ", ".join([str(x) for x in self.StringMap])))

    def write(self, backend, path_to_file, base_name):
        """ Write to file as needed for given backend
        """

        # Bail out if an unsupported backend was passed
        backend_types = self.Types.get("BackendTypes", {})
        if backend.Type not in backend_types:
            print(
                "** ERROR: argMultiFontStringHelper cannot use backend of type {} to write to file. Ignoring it.".format(
                    backend.Type))
            return

        # Delegate string creation to backend
        caption_string = self.Backend.generate_multi_font_string(self)

        # Write to file with appropriate extension for given backend
        caption_file_name = "{}.{}".format(
            base_name,
            backend_types.get(backend.Type, {}).get("captions", ''))
        with open(os.path.join(path_to_file, caption_file_name),
                  'w') as f:
            f.write("%s" % caption_string)

    def append(self, string, font, color=None):
        """ Try to append string/font pair to internal string
        """

        # Bail out if a non-string type was passed as first input
        if not isinstance(string, str):
            print(
                "*  WARNING: attempted to append {} instead of string to argMultiFontStringHelper. Ignoring it.".format(
                    type(string)))
            return

        # Treat unrecognized font types as default
        self.StringMap.append((
            string,
            self.Types.get("FontTypes", {}).get(font, 0),
            color))

    def iterator(self):
        """ Provide iterator over internals
        """

        # Create generator over list of doublets
        for (string, font_bytes, color) in self.StringMap:
            yield string, font_bytes, color

    def execute_backend(self, handle=None):
        """ Delegate artifact creation to backend
        """

        # Create generator over list of doublets
        return self.Backend.generate_multi_font_string(self, handle)



#HEADER
#                          arg/Tools/Utilities.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

import collections
import os
import sys

import yaml


def update(dict1, dict2):
    """ Recursively update first dict with second dict entries
    """

    for k, v in dict2.items():
        if isinstance(v, collections.Mapping):
            dict1[k] = update(dict1.get(k, {}), v)
        else:
            dict1[k] = v
    return dict1


def read_yml_file(file, blocking=False):
    """ Check extension and read YML file
    """

    # Check whether file exists
    if file:
        res = os.path.splitext(file)
        extension = res[-1]
    else:
        errStr = "no file to be read."
        extension = ''

    # Ensure that extension is understood
    if extension in (".yml", ".yaml", ".arg"):
        if os.path.exists(file.strip()):
            if os.path.isfile(file.strip()):
                try:
                    return yaml.safe_load(open(file))
                except yaml.MarkedYAMLError as e:
                    print("** ERROR: invalid YAML file {} in line {} ({} {}). Exiting.".format(
                        file,
                        e.problem_mark.line,
                        e.problem,
                        e.context))
                    sys.exit(1)
            else:
                errStr = "** ERROR: Specified file {} is not a file.".format(file)
        # Error out if file does not exist
        else:
            errStr = "specified file {} does not exist.".format(file)
    # Error out if format is not supported
    else:
        errStr = "file type {} not supported in {}.".format(extension, file)

    # Exit early if incorrect yml file is blocking
    if blocking and errStr:
        print("** ERROR: {} Exiting.".format(
            errStr.capitalize()))
        sys.exit(1)

    # Only return error string otherwise
    else:
        return errStr



#HEADER
#                       arg/GUI/Logic/argActionManager.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER


from PySide2.QtCore import QObject, QSettings
from PySide2.QtWidgets import QApplication

from arg.GUI.Logic.argAboutAction import argAboutAction
from arg.GUI.Logic.argCleanAction import argCleanAction
from arg.GUI.Logic.argHelpAction import argHelpAction
from arg.GUI.Logic.argOpenAction import argOpenAction
from arg.GUI.Logic.argOpenRecentAction import argOpenRecentAction
from arg.GUI.Logic.argOpenUserSettingsAction import argOpenUserSettingsAction
from arg.GUI.Logic.argQuitAction import argQuitAction
from arg.GUI.Logic.argReloadAction import argReloadAction
from arg.GUI.Logic.argRunAction import argRunAction
from arg.GUI.Logic.argSaveAction import argSaveAction
from arg.GUI.Logic.argSaveAsAction import argSaveAsAction
from arg.GUI.Logic.argSaveBeforeRunAction import argSaveBeforeRunAction


class argActionManager(QObject):
    """An action manager class to cover all available actions
    """

    def __init__(self):
        super().__init__()

        # All available actions
        self.quitAct = argQuitAction()
        self.openAct = argOpenAction()
        self.saveAct = argSaveAction()
        self.saveAsAct = argSaveAsAction()
        self.saveBeforeRunAct = argSaveBeforeRunAction()
        self.runAct = argRunAction()
        self.cleanAct = argCleanAction()
        self.reloadAct = argReloadAction()
        self.helpAct = argHelpAction()
        self.openUserSettingsAct = argOpenUserSettingsAction()
        self.aboutAct = argAboutAction()
        self.maxNumberOfActions = 10
        self.openRecentActs = []

    def initActionFromSettings(self):

        # Fill actions from settings
        settings = QApplication.instance().settingsController
        permanentSettings = QSettings(QSettings.IniFormat, QSettings.UserScope, settings.getCompanyName(),
                                      settings.getToolName())
        recentMenuFilesNames = permanentSettings.value(settings.getRecentMenuSettings())
        if recentMenuFilesNames:
            if isinstance(recentMenuFilesNames, str):
                self.createOpenRecentAction(recentMenuFilesNames)
            else:
                # To keep the order of the list with insertion comment, we need to inverse it
                recentMenuFilesNamesInverted = []
                for fileName in recentMenuFilesNames:
                    recentMenuFilesNamesInverted.insert(0, fileName)
                for fileName in recentMenuFilesNamesInverted:
                    self.createOpenRecentAction(fileName)

    def createOpenRecentAction(self, fileName):

        # Verify if the fileName already exists
        needsCreate = True
        for openRecentAct in self.openRecentActs:
            if openRecentAct.getFileName() == fileName:
                # If the action already exists, move it at the top of the list and return it
                self.openRecentActs.remove(openRecentAct)
                self.openRecentActs.insert(0, openRecentAct)
                needsCreate = False

        if needsCreate:
            openRecentAction = argOpenRecentAction(fileName, self)
            self.openRecentActs.insert(0, openRecentAction)

            # If the number of actions is more than the maximum allowed, we remove the old one
            if len(self.openRecentActs) > self.maxNumberOfActions:
                self.openRecentActs.pop(len(self.openRecentActs) - 1)

        # Settings
        settings = QApplication.instance().settingsController
        permanentSettings = QSettings(QSettings.IniFormat, QSettings.UserScope, settings.getCompanyName(),
                                      settings.getToolName())

        openRecentActSttings = []
        for openRecentAct in self.openRecentActs:
            openRecentActSttings.append(openRecentAct.getFileName())

        permanentSettings.setValue(settings.getRecentMenuSettings(), openRecentActSttings)

        if needsCreate:
            return openRecentAction
        else:
            return None

    def getActions(self):

        return self.openRecentActs



#HEADER
#                         arg/GUI/View/argMainWindow.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

from PySide2.QtWidgets import QApplication, QMainWindow, QToolBar, QRadioButton, QWidget, QSizePolicy

from arg.GUI.View.argCentralWidget import argCentralWidget


class argMainWindow(QMainWindow):
    """A widget class
    """

    def __init__(self):
        super().__init__()
        qapp = QApplication.instance()

        # 'File' menu section
        fileMenu = self.menuBar().addMenu("&File")
        fileMenu.addAction(qapp.actionManager.openAct)
        self.recentMenu = fileMenu.addMenu("&Open Recent")
        fileMenu.addSeparator()
        fileMenu.addAction(qapp.actionManager.saveAct)
        fileMenu.addAction(qapp.actionManager.saveAsAct)
        fileMenu.addSeparator()
        fileMenu.addAction(qapp.actionManager.reloadAct)
        fileMenu.addSeparator()
        fileMenu.addAction(qapp.actionManager.quitAct)

        # 'ARG' menu section
        argMenu = self.menuBar().addMenu("&ARG")
        argMenu.addAction(qapp.actionManager.runAct)
        argMenu.addAction(qapp.actionManager.cleanAct)
        argMenu.addSeparator()
        argMenu.addAction(qapp.actionManager.saveBeforeRunAct)
        argMenu.addSeparator()
        argMenu.addAction(qapp.actionManager.openUserSettingsAct)

        # '?' menu section
        questionMenu = self.menuBar().addMenu("&?")
        questionMenu.addAction(qapp.actionManager.helpAct)
        questionMenu.addAction(qapp.actionManager.aboutAct)

        # Run settings widgets
        runnerERadioButton = QRadioButton("Run ARG with Explorator", self)
        runnerERadioButton.setChecked(True)
        runnerGRadioButton = QRadioButton("Run ARG with Generator", self)

        # Spacer to align help widget to the right
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        spacer.setVisible(True)

        # Toolbar widgets
        toolbar = QToolBar(self)
        toolbar.addAction(qapp.actionManager.openAct)
        toolbar.addAction(qapp.actionManager.reloadAct)
        toolbar.addAction(qapp.actionManager.saveAct)
        toolbar.addSeparator()
        toolbar.addAction(qapp.actionManager.cleanAct)
        toolbar.addAction(qapp.actionManager.runAct)
        toolbar.addWidget(runnerERadioButton)
        toolbar.addWidget(runnerGRadioButton)
        toolbar.addSeparator()
        toolbar.addWidget(spacer)
        toolbar.addSeparator()
        toolbar.addAction(qapp.actionManager.helpAct)
        self.addToolBar(toolbar)

        # Main widgets
        centralWidget = argCentralWidget()
        self.setCentralWidget(centralWidget)

        # Toolbar connections
        runnerERadioButton.clicked.connect(self.onRunEButtonClicked)
        runnerGRadioButton.clicked.connect(self.onRunGButtonClicked)

    def initRecentMenuFromSettings(self):
        """ Init recent file menu from the settings
        """
        qapp = QApplication.instance()
        qapp.actionManager.initActionFromSettings()
        self.recentMenu.clear()
        self.recentMenu.addActions(qapp.actionManager.getActions())

    def addRecentFileOpenAction(self, filename):
        """Fill central widget with provided data
        """

        qapp = QApplication.instance()
        qapp.actionManager.createOpenRecentAction(filename)
        self.recentMenu.clear()
        self.recentMenu.addActions(qapp.actionManager.getActions())

    def fillParameters(self, data):
        """Fill central widget with provided data
        """

        self.centralWidget().fillParameters(data)

    def constructParameters(self):
        """Initialize central widget and construct corresponding data
        """

        return self.centralWidget().constructParameters()

    def logStandard(self, log):
        """Fill central widget with provided standard log message
        """

        self.centralWidget().logStandard(log)

    def logError(self, log):
        """Fill central widget with provided error log message
        """

        self.centralWidget().logError(log)

    def clearLogs(self):
        """clear log from central widget
        """

        self.centralWidget().clearLogs()

    def onRunEButtonClicked(self):
        """ Switch runner to E execution mode
        """

        qapp = QApplication.instance()
        qapp.runEButtonClicked()
        self.centralWidget().runEButtonClicked()

    def onRunGButtonClicked(self):
        """ Switch runner to G execution mode
        """
        qapp = QApplication.instance()
        qapp.runGButtonClicked()
        self.centralWidget().runGButtonClicked()



#HEADER
#                      arg/GUI/Logic/argParameterWriter.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

import yaml
from PySide2.QtCore import QObject

from arg.GUI.Logic.argParameterData import argParameterData
from arg.GUI.Logic.argSettingsController import argSettingsController


class argParameterWriter(QObject):
    """A writer class to write parameters file
    """

    def __init__(self, data=None):
        super().__init__()

        if data is None:
            data = dict()

        self.ParameterData = argParameterData(data)
        self.SettingController = argSettingsController()

    def setData(self, data):
        """Data setter
        """

        self.ParameterData = argParameterData(data)

    def write(self, filePath):
        """Dump current parameters data into specified file
        """

        # Retrieve current widget values in all tabs

        # 'Report Information' tab
        reportInformationToWrite = {}
        if self.SettingController.backendTypeKey() in self.ParameterData.Dict:
            reportInformationToWrite[
                self.SettingController.backendTypeParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.backendTypeKey())
        if self.SettingController.reportTypeKey() in self.ParameterData.Dict:
            reportInformationToWrite[self.SettingController.reportTypeParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.reportTypeKey())
        if self.SettingController.classificationKey() in self.ParameterData.Dict:
            reportInformationToWrite[
                self.SettingController.classificationParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.classificationKey())
        if self.SettingController.fileNameKey() in self.ParameterData.Dict:
            reportInformationToWrite[self.SettingController.fileNameParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.fileNameKey())
        if self.SettingController.mutablesFileKey() in self.ParameterData.Dict:
            mutablesFile = self.ParameterData.Dict.get(
                self.SettingController.mutablesFileKey())
            reportInformationToWrite[self.SettingController.mutablesFileParameterFileKey()] = \
                mutablesFile if not mutablesFile or ".yml" in mutablesFile or ".yaml" in mutablesFile \
                    else "{}.yml".format(mutablesFile)
        if self.SettingController.structureFileKey() in self.ParameterData.Dict:
            structureFile = self.ParameterData.Dict.get(
                self.SettingController.structureFileKey())
            reportInformationToWrite[self.SettingController.structureFileParameterFileKey()] = \
                structureFile if not structureFile or ".yml" in structureFile or ".yaml" in structureFile \
                    else "{}.yml".format(structureFile)
        if self.SettingController.structureEndFileKey() in self.ParameterData.Dict:
            structureEndFile = self.ParameterData.Dict.get(
                self.SettingController.structureEndFileKey())
            reportInformationToWrite[self.SettingController.structureEndFileParameterFileKey()] = \
                structureEndFile if not structureEndFile or ".yml" in structureEndFile or ".yaml" in structureEndFile \
                    else "{}.yml".format(structureEndFile)
        if self.SettingController.artifactFileKey() in self.ParameterData.Dict:
            reportInformationToWrite[
                self.SettingController.artifactFileParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.artifactFileKey())
        if self.SettingController.outputFolderKey() in self.ParameterData.Dict:
            reportInformationToWrite[
                self.SettingController.outputFolderParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.outputFolderKey())

        if self.SettingController.verbosityKey() in self.ParameterData.Dict:
            reportInformationToWrite[self.SettingController.verbosityParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.verbosityKey())

        # 'General Options' tab
        generalOptionsToWrite = {}
        if self.SettingController.titleKey() in self.ParameterData.Dict:
            generalOptionsToWrite[self.SettingController.titleParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.titleKey())
        if self.SettingController.numberKey() in self.ParameterData.Dict:
            generalOptionsToWrite[self.SettingController.numberParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.numberKey())
        if self.SettingController.issueKey() in self.ParameterData.Dict:
            generalOptionsToWrite[self.SettingController.issueParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.issueKey())
        if self.SettingController.versionsKey() in self.ParameterData.Dict:
            generalOptionsToWrite[self.SettingController.versionParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.versionsKey())
        if self.SettingController.authorsKey() in self.ParameterData.Dict:
            generalOptionsToWrite[self.SettingController.authorParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.authorsKey())
        if self.SettingController.organizationsKey() in self.ParameterData.Dict:
            generalOptionsToWrite[self.SettingController.organizationParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.organizationsKey())
        if self.SettingController.locationKey() in self.ParameterData.Dict:
            generalOptionsToWrite[self.SettingController.locationParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.locationKey())
        if self.SettingController.yearKey() in self.ParameterData.Dict:
            generalOptionsToWrite[self.SettingController.yearParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.yearKey())
        if self.SettingController.monthKey() in self.ParameterData.Dict:
            generalOptionsToWrite[self.SettingController.monthParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.monthKey())
        if self.SettingController.abstractFileKey() in self.ParameterData.Dict:
            generalOptionsToWrite[self.SettingController.abstractFileParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.abstractFileKey())
        if self.SettingController.prefaceKey() in self.ParameterData.Dict:
            generalOptionsToWrite[self.SettingController.prefaceParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.prefaceKey())
        if self.SettingController.thanksKey() in self.ParameterData.Dict:
            generalOptionsToWrite[self.SettingController.thanksParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.thanksKey())
        if self.SettingController.executiveSummaryKey() in self.ParameterData.Dict:
            generalOptionsToWrite[
                self.SettingController.executiveSummaryParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.executiveSummaryKey())
        if self.SettingController.nomenclatureKey() in self.ParameterData.Dict:
            generalOptionsToWrite[self.SettingController.nomenclatureParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.nomenclatureKey())
        if self.SettingController.finalKey() in self.ParameterData.Dict:
            generalOptionsToWrite[self.SettingController.finalParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.finalKey())
        if self.SettingController.keySeparatorKey() in self.ParameterData.Dict:
            generalOptionsToWrite[self.SettingController.keySeparatorParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.keySeparatorKey())

        # 'Data' tab
        dataToWrite = {}
        if self.SettingController.dataFolderKey() in self.ParameterData.Dict:
            dataToWrite[self.SettingController.dataFolderParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.dataFolderKey())
        if self.SettingController.geometryRootKey() in self.ParameterData.Dict:
            dataToWrite[self.SettingController.geometryRootParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.geometryRootKey())
        if self.SettingController.reportedCadMetaDataKey() in self.ParameterData.Dict:
            dataToWrite[self.SettingController.reportedCadMetaDataParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.reportedCadMetaDataKey())
        if self.SettingController.inputDeckKey() in self.ParameterData.Dict:
            inputDeck = self.ParameterData.Dict.get(self.SettingController.inputDeckKey())
            dataToWrite[self.SettingController.inputDeckParameterFileKey()] = inputDeck
        if self.SettingController.ignoredBlocksKey() in self.ParameterData.Dict:
            dataToWrite[self.SettingController.ignoredBlocksParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.ignoredBlocksKey())
        # Mappings
        if self.SettingController.bijectiveMappingsKey() in self.ParameterData.Dict:
            dataToWrite[self.SettingController.bijectiveMappingsParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.bijectiveMappingsKey())

        if self.SettingController.solutionCasesKey() in self.ParameterData.Dict:
            dataToWrite[self.SettingController.solutionCasesParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.solutionCasesKey())

        # 'Inserts' tab
        insertsToWrite = {}
        if self.SettingController.insertionKey() in self.ParameterData.Dict:
            insertsToWrite[self.SettingController.insertionParameterFileKey()] = self.ParameterData.Dict.get(
                self.SettingController.insertionKey())

        # Dump all retrieved data dicts into specified file
        with open(filePath, 'w') as file:
            if len(reportInformationToWrite.keys()) > 0:
                yaml.dump(reportInformationToWrite, file)
            if len(generalOptionsToWrite.keys()) > 0:
                yaml.dump(generalOptionsToWrite, file)
            if len(dataToWrite.keys()) > 0:
                yaml.dump(dataToWrite, file)
            if len(insertsToWrite.keys()) > 0:
                yaml.dump(insertsToWrite, file)



#HEADER
#                      arg/Backend/argWordBackend.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

import numbers
import os
import platform
import re
import subprocess

import clr
import yaml
from docx.enum.text import WD_PARAGRAPH_ALIGNMENT as WD_ALIGN_PARAGRAPH
from docx.oxml import OxmlElement
from docx.oxml.ns import qn
from docx.shared import Mm, Cm, Inches, RGBColor, Pt
from pylatex import NoEscape

from arg.Common.argMultiFontStringHelper import argMultiFontStringHelper
from arg.Backend.argBackendBase import argBackendBase

osn = os.name

# Try importing Word Application in Python environment
try:
    clr.AddReference("System")
    clr.AddReference("Microsoft.Office.Interop")
    from Microsoft.Office.Interop import Word

# Find work-around
except:
    try:
        # Use Windows32 client on Windows only
        if osn == "nt":
            from win32com.client import DispatchEx
    except:
        pass

# Load supported colors
common_dir = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(common_dir, "../Common/argTypes.yml"),
          'r',
          encoding="utf-8") as t_file:
    supported_types = yaml.safe_load(t_file)

# Retrieve supported colors
colors = supported_types.get("FontColors")

# Retrieve supported fonts
fonts = supported_types.get("BackendTypes").get("Word").get("Fonts")


class argWordBackend(argBackendBase):
    """A concrete class providing a Word backend
    """

    # Map from greek letters to their respective unicodes
    GreekLetters = {
        "Alpha": u'\u0391',
        "Beta": u'\u0392',
        "Gamma": u'\u0393',
        "Delta": u'\u0394',
        "Epsilon": u'\u0395',
        "Zeta": u'\u0396',
        "Eta": u'\u0397',
        "Theta": u'\u0398',
        "Iota": u'\u0399',
        "Kappa": u'\u039A',
        "Lamda": u'\u039B',
        "Mu": u'\u039C',
        "Nu": u'\u039D',
        "Xi": u'\u039E',
        "Omicron": u'\u039F',
        "Pi": u'\u03A0',
        "Rho": u'\u03A1',
        "Sigma": u'\u03A3',
        "Tau": u'\u03A4',
        "Upsilon": u'\u03A5',
        "Phi": u'\u03A6',
        "Chi": u'\u03A7',
        "Psi": u'\u03A8',
        "Omega": u'\u03A9',
        "alpha": u'\u03B1',
        "beta": u'\u03B2',
        "gamma": u'\u03B3',
        "delta": u'\u03B4',
        "epsilon": u'\u03B5',
        "zeta": u'\u03B6',
        "eta": u'\u03B7',
        "theta": u'\u03B8',
        "iota": u'\u03B9',
        "kappa": u'\u03BA',
        "lamda": u'\u03BB',
        "mu": u'\u03BC',
        "nu": u'\u03BD',
        "xi": u'\u03BE',
        "omicron": u'\u03BF',
        "pi": u'\u03C0',
        "rho": u'\u03C1',
        "sigma": u'\u03C3',
        "tau": u'\u03C4',
        "upsilon": u'\u03C5',
        "phi": u'\u03C6',
        "chi": u'\u03C7',
        "psi": u'\u03C8',
        "omega": u'\u03C9'}

    def __init__(self, parameters=None):

        # Call superclass init
        super().__init__(parameters)

        # Define backend type
        self.Type = "Word"

        self.colors = None
        self.fonts = None
        self.app = "argWordBackend"
        self.osn = os.name

        # Try importing Word Application in Python environment
        try:
            clr.AddReference("System")
            clr.AddReference("Microsoft.Office.Interop")
            from Microsoft.Office.Interop import Word

        # Find work-around
        except:
            try:
                # Use Windows32 client on Windows only
                if self.osn == "nt":
                    from win32com.client import DispatchEx
            except:
                pass
        self.__init_fonts_colors()

    def __init_fonts_colors(self):
        # Load supported colors
        common_dir = os.path.dirname(os.path.realpath(__file__))
        with open(os.path.join(common_dir, "../Common/argTypes.yml"),
                  'r',
                  encoding="utf-8") as t_file:
            supported_types = yaml.safe_load(t_file)

        # Retrieve supported colors
        self.colors = supported_types.get("FontColors")

        # Retrieve supported fonts
        self.fonts = supported_types.get("BackendTypes").get("Word").get("Fonts")

    @staticmethod
    def generate_text(text, type="default"):
        """Add page break to the report
        """

        # Generate a backend dependent text
        return "{}".format(text)

    def generate_multi_font_string(self, multi_font_string, paragraph=None):
        """Either append Word runs to paragraph or return text string
        """

        # Append Word runs to paragraph when one was passed
        if paragraph:

            # Iterate over all items in string helper
            for string, font_bits, color in multi_font_string.iterator():

                # Map Greek letter to unicode when requested
                if font_bits & 16 == 16:
                    string = self.GreekLetters.get(string, '')

                # Add a run per string and set font per defined bits
                run = paragraph.add_run(string)
                run.italic = font_bits & 1 == 1
                run.bold = font_bits & 2 == 2
                if font_bits & 4 == 4:
                    run.font.name = self.fonts.get("typewriter", "default" if self.fonts.get("default") else None)
                elif font_bits & 8 == 8:
                    run.font.name = self.fonts.get("calligraphic", "default" if self.fonts.get("default") else None)
                elif self.fonts.get("default", None):
                    run.font.name = self.fonts.get("default", None)
                if color:
                    rgb = self.colors.get(color).split(',')
                    if len(rgb) > 2:
                        run.font.color.rgb = RGBColor(*[int(x) for x in rgb[:3]])

            # Nothing is returned in this case
            return

        # Return a plain string ignoring font specifications otherwise
        else:
            return ''.join([x[0] for x in multi_font_string.StringMap])

    def generate_matrix_string(self, matrix):
        """Generate Word string from matrix entries
        """

        # Bail out early if vector is empty or does not have correct type
        if not matrix or not isinstance(matrix, (list, tuple)):
            return ''

        # Return plain text string
        return ','.join(["[{}]".format(','.join([f"{x}" for x in v])) for v in matrix])

    @staticmethod
    def create_list(paragraph, list_type):
        """Create list
        """

        # Retrieve XML paragraph element
        p = paragraph._p

        # Retrieve paragraph properties
        pPr = p.get_or_add_pPr()

        # Create number properties element
        numPr = OxmlElement('w:numPr')

        # Create numId element to set bullet type
        numId = OxmlElement('w:numId')

        # Set list type/indentation
        numId.set(qn('w:val'), list_type)

        # Add bullet type to number properties list
        numPr.append(numId)

        # Add number properties to paragraph
        pPr.append(numPr)


    def add_list(self, item, num_lvl=1, number_items=False):
        """Add itemization or enumeration to the report
        """

        # Iterate over all elements of provided list
        for it in item['items']:
            # Add string elements
            if "string" in it.keys():
                self.Report.add_paragraph(it["string"])

            # Add numbered elements
            if "number" in it.keys():
                self.Report.add_paragraph(it["number"])


    def add_comment(self,  comments_dict, key):
        """Add comment to the report from a dict as either text string
        or as sub-paragraph depending on number of dict values (1 or 2)
        """

        # Retrieve comment for given key and bail out if none found
        comment = comments_dict.get(str(key))
        if not comment:
            return False

        # Insert either sub-paragraph or string
        if len(comment) > 1:
            self.Report.add_paragraph(
                comment[0] + ":", "Subtitle")
            self.Report.add_paragraph(comment[1])
        else:
            self.Report.add_paragraph(comment[0])

        # Comment was found and inserted
        return True


    def add_paragraph(self, item, p=None):
        """Add paragraph to the report
        """

        # Check whether a string or a file is to be included
        if "string" in item:
            # Retrieve provided string
            string = item["string"]

            # Distinguish between plain and multi-font strings
            if isinstance(string, argMultiFontStringHelper):
                # Insert string into the report
                self.generate_multi_font_string(
                    string, self.Report.add_paragraph())
            else:
                # Directly insert Word fragment into the report
                self.Report.add_paragraph(string)

        # Insert verbatim fragment
        elif "verbatim" in item:
            # Create paragraph in macro style
            p = self.Report.add_paragraph(item["verbatim"])
            p.style = self.Report.styles["macro"]

        # Insert contents of a file
        elif "include" in item and len(item["include"]) > 0:
            curr_item = item["include"][0]
            if "file" in curr_item:
                file_name = os.path.join(self.Parameters.DataDir, curr_item["file"])
                font = curr_item.get("font")
                size = curr_item.get("size")
                italic = curr_item.get("italic", False)
                underline = curr_item.get("underline", False)
                color = curr_item.get("color")
                alignment = curr_item.get("alignment")
                left_indent = curr_item.get("left_indent")
                right_indent = curr_item.get("right_indent")
                first_line_indent = curr_item.get("first_line_indent")
                space_before = curr_item.get("space_before")
                space_after = curr_item.get("space_after")
                line_spacing = curr_item.get("line_spacing")
                keep_with_next = curr_item.get("keep_with_next")
            else:
                file_name = os.path.join(self.Parameters.DataDir, item["include"])
                font, size, color = None, None, None
                italic, underline = False, False
                alignment = None
                left_indent, right_indent, first_line_indent = None, None, None
                space_before, space_after = None, None
                line_spacing = None
                keep_with_next = None

            with open(file_name, 'r') as tex_file:
                p = self.Report.add_paragraph()
                run = p.add_run()

                # Read each row and add it to Word output
                for line in tex_file:
                    run.text = run.text + line

                # Add font commands if provided
                if font == "typewriter":
                    run.font.name = self.fonts.get(font, "default" if self.fonts.get("default") else None)
                elif font == "calligraphic":
                    run.font.name = self.fonts.get(font, "default" if self.fonts.get("default") else None)
                if size:
                    run.font.size = Pt(size)
                if italic:
                    run.font.italic = italic
                if underline:
                    run.font.underline = underline
                if color:
                    rgb = color.split(',')
                    if len(rgb) > 2:
                        run.font.color.rgb = RGBColor(*[int(x) for x in rgb[:3]])
                if alignment:
                    p.alignment = WD_ALIGN_PARAGRAPH.CENTER
                if left_indent:
                    p.left_indent = Pt(left_indent)
                if right_indent:
                    run.right_indent = Pt(right_indent)
                if first_line_indent:
                    run.first_line_indent = Pt(first_line_indent)
                if space_before:
                    run.space_before = Pt(space_before)
                if space_after:
                    run.space_after = Pt(space_after)
                if line_spacing:
                    run.line_spacing = Pt(line_spacing)
                if keep_with_next:
                    run.keep_with_next = keep_with_next

    def add_subdivision(self, item, level):
        """Add specified subdivision to the report
        """

        # Retrieve title string
        title_string = item.get("title", '')

        # Handle title string according to its type
        if isinstance(title_string, str):
            # Directly insert base strings
            self.Report.add_heading(title_string, level)

        elif isinstance(title_string, argMultiFontStringHelper):
            # Handle multi-font string helpers
            self.generate_multi_font_string(
                title_string,
                self.Report.add_heading())

        # Add subdivision paragraph header if provided
        self.add_paragraph(item)

    def add_subtitle(self, item):
        """Add subsection to the report
        """

        # Call appropriate subdivision method
        self.add_subdivision(item, 3)

    def add_subsection(self, item, numbered=True):
        """Add subsection to the report
        """

        # Call appropriate subdivision method
        self.add_subdivision(item, 3)

    def add_section(self, item, numbered=True):
        """Add section to the report
        """

        # Call appropriate subdivision method
        self.add_subdivision(item, 2)

    def add_chapter(self, item, numbered=True):
        """Add chapter to the report
        """

        # Start with new page
        self.Report.add_page_break()

        # Call appropriate subdivision method
        self.add_subdivision(item, 1)

    def add_page_break(self):
        """Add page break to the report
        """

        # Use eponymous command from Word backend
        self.Report.add_page_break()

    def add_heading(self, text, lvl=1, p=None):
        """Add heading to the report
        """

        # Use provided paragraph
        if p:
            p.text = text

        # Otherwise, use eponymous command from Word backend
        else:
            self.Report.add_heading(text, lvl)

    def add_table(self, header_list, body, caption_string, do_verbatim=False, _=None):
        """Add table to the report
        """

        # Bail out early if header is empty
        try:
            n_cols = len(header_list)
            if not n_cols:
                return
        except:
            return

        # Bail out early if unsupported body type
        if not body:
            return
        if type(body) == dict:
            is_dict = True
        elif type(body) == list:
            is_dict = False
        else:
            return

        # Create table with required number of columns
        table = self.Report.add_table(
            rows=1,
            cols=n_cols,
            style=self.Report.styles["Light List Accent 1"])
        table.alignment = WD_ALIGN_PARAGRAPH.CENTER

        # Create table header
        cells = table.rows[0].cells
        for c, header_string in zip(cells, header_list):
            # Fill header cell
            if isinstance(header_string, str):
                # Directly insert base strings
                c.text = header_string

            elif isinstance(header_string, argMultiFontStringHelper):
                # Handle multi-font string helpers
                self.generate_multi_font_string(
                    header_string,
                    c.paragraphs[0])

        # Iterate over of contents to create table body
        if is_dict:
            # Iterate over dictionary entries
            for k, v in sorted(body.items()):
                # Add a row of cells and fill those
                cells = table.add_row().cells

                # Make key and values lists if they are strings or numbers
                if isinstance(k, (str, numbers.Number)):
                    k = [k]
                if isinstance(v, (str, numbers.Number)):
                    v = [v]

                # Iterate to fill all cells
                for c, cell_string in zip(cells, list(k) + list(v)):
                    # Distinguish argMultiFontStrings from other types
                    if isinstance(cell_string, argMultiFontStringHelper):
                        # Handle multi-font string helpers
                        self.generate_multi_font_string(
                            "{}".format(cell_string),
                            c.paragraphs[0])
                    else:
                        # Directly insert base strings
                        c.text = "{}".format(cell_string)

                    # Use verbatim style when requested
                    if do_verbatim:
                        c.paragraphs[0].style = self.Report.styles["macro"]
        else:
            # Iterate over list entries
            for row in body:
                # Add a row of cells and fill those
                cells = table.add_row().cells
                for c, cell_string in zip(cells, row):
                    if isinstance(cell_string, str):
                        # Directly insert base strings
                        c.text = cell_string

                    elif isinstance(cell_string, argMultiFontStringHelper):
                        # Handle multi-font string helpers
                        self.generate_multi_font_string(
                            cell_string,
                            c.paragraphs[0])

                    # Use verbatim style when requested
                    if do_verbatim:
                        c.paragraphs[0].style = self.Report.styles["macro"]

        # Create table caption depending on its type
        if caption_string:
            if isinstance(caption_string, str):
                # Directly insert base strings
                caption = self.add_caption("Table", caption_string)
                caption.alignment = WD_ALIGN_PARAGRAPH.CENTER

            elif isinstance(caption_string, argMultiFontStringHelper):
                # Handle multi-font string helpers
                caption = self.add_caption("Table", "")
                caption.alignment = WD_ALIGN_PARAGRAPH.CENTER

                # Append Word runs created from multi-font string
                caption_string.execute_backend(caption)

            else:
                # Insert blank paragraph to ensure table is closed
                self.Report.add_paragraph()
        else:
            # Insert blank paragraph to ensure table is closed
            self.Report.add_paragraph()

    def add_type_caption(self, caption_type, caption_string):
        """Add table caption
        """

        # Add basestring type captions in one step
        if isinstance(caption_string, str):
            # Directly insert base strings
            c = self.add_caption("{} ".format(caption_type), caption_string)
            c.alignment = WD_ALIGN_PARAGRAPH.CENTER

        # Add multi front captions in two steps
        elif isinstance(caption_string, argMultiFontStringHelper):
            # Initiate caption with caption type
            c = self.add_caption("{} ".format(caption_type), "")

            # Insert multi-font string into the report
            self.generate_multi_font_string(
                caption_string, c)
            c.alignment = WD_ALIGN_PARAGRAPH.CENTER

            # Append Word runs created from multi-font string
            caption_string.execute_backend()


    def add_figure(self, arguments):
        """Add figure to report
        """

        # A figure width is required
        try:
            figure_width = arguments["width"]
        except KeyError:
            print("*  WARNING: no figure width was provided.")
            return

        # Only float numbers of centimeters or inches are understood
        try:
            width_value, width_unit = float(figure_width[:-2]), figure_width[-2:].lower()
        except:
            print("*  WARNING: could not interpret {} as figure width.".format(
                figure_width))
            return
        if width_unit == "mm":
            actual_width = Mm(width_value)
        elif width_unit == "cm":
            actual_width = Cm(width_value)
        elif width_unit == "in":
            actual_width = Inches(width_value)
        else:
            print("*  WARNING: unknown width unit: {}.".format(
                width_unit))
            return

        # Retrieve image and associated caption for figure creation
        figure_file_name, caption_string = self.fetch_image_and_caption(
            arguments)

        # Create figure when available
        if figure_file_name:
            self.Report.add_picture(
                os.path.join(self.Parameters.OutputDir, figure_file_name),
                width=actual_width)
            p = self.Report.paragraphs[-1]
            p.alignment = WD_ALIGN_PARAGRAPH.CENTER

            # Append caption when available
            if caption_string:
                self.add_type_caption("Figure", caption_string)


    def add_caption(self, caption_type, caption_string):
        """ Create caption area in XML, with proper number
        """

        # Create new paragraph to populate
        paragraph = self.Report.add_paragraph(caption_type.capitalize().strip() + ' ', style="Caption")
        r = paragraph.add_run()._r

        # Create begin XML element
        beginXml = OxmlElement("w:fldChar")
        beginXml.set(qn("w:fldCharType"), "begin")
        beginXml.set(qn("w:dirty"), "true")

        # Create end XML element
        endXml = OxmlElement("w:fldChar")
        endXml.set(qn("w:fldCharType"), "end")

        # Create number XML element
        numberXml = OxmlElement("w:instrText")
        numberXml.text = "SEQ " + caption_type.upper() + " \\* ARABIC"

        # Add all elements to current run
        r.append(beginXml)
        r.append(numberXml)
        r.append(endXml)

        # Add caption string run
        paragraph.add_run(': ' + caption_string)

        return paragraph


    def recursively_build_report(self, item_tree, list_lvl=1, p=None):
        """Recursively traverse structure tree to build Word report
        """

        # Iterate over each dictionary in the array
        for item in item_tree:
            item_type = item.get('n')
            if item_type == "string":
                # Directly insert text fragment into the report
                self.Report.append(NoEscape(item.get("string")))

            # Handle chapter case
            elif item_type.startswith("chapter"):
                # Create chapter
                self.add_chapter(item)

            # Handle section case
            elif item_type.startswith("section"):
                # Create section
                self.add_section(item)

            # Handle subsection case
            elif item_type.startswith("subsection"):
                # Create subsection
                self.add_subsection(item)

            # Handle paragraph case
            elif item_type == "paragraph":
                # Add paragraph to the report
                self.add_paragraph(item)

            # Handle itemization/enumeration case
            elif item_type in ("itemize", "enumerate"):
                # Add itemization/enumeration to the report
                self.add_list(item, list_lvl, True if item_type == "enumerate" else False)
                list_lvl = list_lvl + 1

            # Handle verbatim case
            elif item_type == "verbatim":
                # Include file verbatim as report paragraph
                file_name = os.path.join(self.Parameters.DataDir,
                                         item["include"])
                verbatim_style = self.Report.styles["macro"]
                with open(file_name, 'r') as tex_file:

                    # Read each row and add it to Word output
                    for line in tex_file:
                        self.Report.add_paragraph(line).style = verbatim_style

            # Handle figure case
            elif item_type == "figure":
                # Add figure to the report
                self.add_figure(item["arguments"])

            # Handle meta-information case
            elif item_type == "meta":
                # Append meta-information to the report
                self.add_meta_information(item)

            # Handle properties case
            elif item_type == "properties":
                # Append information specific to given property to the report
                self.add_information(item)

            # Handle aggregate case
            elif item_type == "aggregate":
                # Append aggregated information
                self.add_aggregation(item)

            # Proceed with recursion if needed
            if "sections" in item:
                self.recursively_build_report(item["sections"], list_lvl)


    def assemble(self, report_map, version=None, latex_proc=None):
        """Create a Word report from given report map
        """

        # Call super method first
        super().assemble(report_map, version)

        # Create report with preamble
        print("[argWordBackend] Generating Word report")
        self.create_document_preamble(version)

        # Pass over data for report contents
        self.recursively_build_report(report_map["chapters"], list_lvl=1)

        # Append backend-specific postamble
        self.append_document_postamble()

        # Handle title numbering
        self.add_header_numbering()

        # Create Word report
        print("[argWordBackend] Saving Word report")
        self.Report.save("{}.docx".format(os.path.join(
            self.Parameters.OutputDir,
            self.Parameters.FileName)))

        # Update tables of contents, figures, tables on Windows only for now
        if self.osn == "nt":
            self.update_tocs(r"{}.docx".format(os.path.abspath(os.path.join(
                self.Parameters.OutputDir,
                self.Parameters.FileName))))
        else:
            print("*  WARNING: Table of contents not automatically generated yet by Word backend on {} systems.".format(
                self.osn))

    def add_document_provenance(self, version=None):
        """Add document provenance information on a new page
        """

        # Start new page
        self.Report.add_page_break()

        # Compute provenance information
        timestamp = r'This document was generated on {} with the Automatic Report Generator (ARG) version "{}" on the {} system {}.'.format(
            self.get_timestamp(),
            version,
            platform.system(),
            platform.node())
        self.Report.add_paragraph(timestamp)

    def add_document_tocs(self):
        """Add document Table of Contents, List of Figures and List of Tables on new pages
        """

        # Start new page
        self.Report.add_page_break()

        # Add tables of contents, figures, tables
        self.add_table_of_contents()
        self.Report.add_page_break()
        self.add_index("Figure")
        self.Report.add_page_break()
        self.add_index("Table")
        self.Report.add_page_break()

    @staticmethod
    def update_tocs(docx_file):
        """Update table of contents, list of figures, list of tables
        """

        # Try opening Word Application
        try:
            word = Word.Application()
        # Find work-around
        except:
            # Try to retrieve MS Word application
            try:
                word = DispatchEx("Word.Application")
            except:
                # Try running C#-built exe solution
                try:
                    subprocess.check_call([r"../../arg/Backend/updateWordFields", docx_file])
                    return

                # Return with warning message
                except:
                    print("*  WARNING: Word report cannot be automatically updated.")
                    return

        # Open the docx file in MS Word
        doc = word.Documents.Open(docx_file)
        word.ActiveDocument.Fields.Update()

        # Update all tables
        if doc.TablesOfContents(1):
            doc.TablesOfContents(1).Update()
        if doc.TablesOfFigures(1):
            doc.TablesOfFigures(1).Update()
        if doc.TablesOfFigures(2):
            doc.TablesOfFigures(2).Update()

        # Close file and exit cleanly
        doc.Close(SaveChanges=True)
        word.Quit()

    def add_table_of_contents(self, heading_run=None, body_run=None, lvls="1-3"):
        """Add table of contents
        """

        # Add a heading in existing run if provided
        if heading_run:
            heading_run.text = "Table of Contents"
        # Otherwise, create a heading
        else:
            self.Report.add_heading("Table of Contents")

        # Use body run if provided
        if body_run:
            run = body_run

        # Otherwise, create a new run
        else:
            paragraph = self.Report.add_paragraph()
            run = paragraph.add_run()

        # Add a beginning character
        fldChar = OxmlElement('w:fldChar')
        fldChar.set(qn('w:fldCharType'), 'begin')

        # Create an interpreted XML text element
        instrText = OxmlElement('w:instrText')
        instrText.set(qn('xml:space'), 'preserve')

        # Add level 1 to level 3 contents
        instrText.text = "TOC \\o {} \\h \\z \\u".format(lvls)  # change 1-3 depending on heading levels you need

        # Add a separate character
        fldChar2 = OxmlElement('w:fldChar')
        fldChar2.set(qn('w:fldCharType'), 'separate')

        # Add an instruction XMl text element to above character
        fldChar3 = OxmlElement('w:t')
        fldChar3.text = "Right-click to update field."
        fldChar2.append(fldChar3)

        # And an ending character
        fldChar4 = OxmlElement('w:fldChar')
        fldChar4.set(qn('w:fldCharType'), 'end')

        # Add all above XML elements to current run
        r_element = run._r
        r_element.append(fldChar)
        r_element.append(instrText)
        r_element.append(fldChar2)
        r_element.append(fldChar4)

    def add_index(self, string, heading_run=None, body_run=None):
        """Add table of contents
        """

        heading = "List of " + string.capitalize().strip() + "s"

        # Add a heading in existing run if provided
        if heading_run:
            heading_run.text = heading

        # Otherwise, create a heading
        else:
            self.Report.add_heading("List of " + string.capitalize().strip() + "s")

        # Use body run if provided
        if body_run:
            run = body_run

        # Otherwise, create a new run
        else:
            paragraph = self.Report.add_paragraph()
            run = paragraph.add_run()

        # Add a beginning character
        fldChar = OxmlElement('w:fldChar')
        fldChar.set(qn('w:fldCharType'), 'begin')

        # Create an interpreted XML text element
        instrText = OxmlElement('w:instrText')
        instrText.set(qn('xml:space'), 'preserve')

        # Add report elements defined by string
        instrText.text = 'TOC \\h \\z \\c "' + string + '" CHARFORMAT'

        # Add an update XML element
        fldChar2 = OxmlElement('w:updateFields')

        # Add an ending character
        fldChar4 = OxmlElement('w:fldChar')
        fldChar4.set(qn('w:fldCharType'), 'end')

        # Add all avobe XML elements to current run
        r_element = run._r
        r_element.append(fldChar)
        r_element.append(fldChar2)
        r_element.append(instrText)
        r_element.append(fldChar4)
        p_element = paragraph._p

    @staticmethod
    def iter_heading(paragraphs):
        """Iterate on headings to number them, a fix suggested by
           https://github.com/python-openxml/python-docx/issues/590
        """

        # Iterate over paragraphs
        for paragraph in paragraphs:
            isItHeading = re.match('Heading ([1-9])', paragraph.style.name)
            # Issue an index only when a heading is encountered
            if isItHeading:
                yield int(isItHeading.groups()[0]), paragraph

    def add_header_numbering(self):
        """Create the sub-tiered header number
        """

        # Allow for up to a maximum of five tiers
        hNums = [0, 0, 0, 0, 0]

        # Iterate over headings
        for index, hx in self.iter_heading(self.Report.paragraphs):
            # Put zeroes in tiers below current one
            for i in range(index + 1, 5):
                hNums[i] = 0

            # Increment current level
            hNums[index] += 1

            # Prepare tiered header index string
            hStr = ''
            for i in range(1, index + 1):
                hStr += "%d." % hNums[i]

            # Add the numbering
            hx.text = "{} {}".format(
                hStr,
                hx.text)



#HEADER
#               arg/DataInterface/argKeyValueFilesReader.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

import os

import h5py

from arg.Common.argInformationObject import argInformationObject
from arg.DataInterface.argDataInterfaceBase import argDataInterfaceBase


class argHDF5Reader(argDataInterfaceBase):
    """A concrete data interface to a serial HDF5 file
    """

    def __init__(self, full_name, _):
        """Default constructor: serial reader only, merge duplicates
        """

        # Instantiate single HDF5 reader
        self.Reader = h5py.File(full_name, 'r')

        # Define visitor to populate internal containers
        def visitor(name, obj):
            if isinstance(obj, h5py.Dataset):
                self.Datasets.update({obj.name: obj.ref})
            else:
                self.Groups.append(obj.name)
            self.Attributes.update({obj.name: list(obj.attrs)})

        # Initialize internal containers and call visitor
        self.Groups = [self.Reader.name]
        self.Datasets = {}
        self.Attributes = {self.Reader.name: list(self.Reader.attrs)}
        self.Reader.visititems(visitor)

    def get_accessors(self):
        """Return singleton of unique HDF5 file name
        """

        return [self.Reader]

    def get_meta_information(self):
        """Retrieve meta-information from data
        """
        # Initialize single reader meta-information
        r_meta = {"file name": os.path.basename(self.Reader.filename), "number of groups": len(self.Groups),
                  "number of datasets": len(self.Datasets),
                  "number of attributes": sum([len(x) for x in self.Attributes.values()])}

        # Return global meta-information
        return [r_meta]

    def get_property_information(self, prop_type, prop_items=None, info_obj=None):
        """Retrieve all information about given sproperty from HDF5 file
        """

        # Bail out early if no property type was provided
        if not prop_type:
            return None

        # Look for property item values when requested and Handle property requests of type group
        if prop_items and prop_type.lower() == "group":
            # Iterate over property items
            for p_item in prop_items:
                # Look for desired items
                if p_item in [os.path.basename(x) for x in self.Groups]:
                    pass

        # Look for matching property type if no speciic items requested
        else:
            # Returned information is dictionary of lists of lists
            info_obj = argInformationObject("arg_dict_lists_lists")

            # Set information names with a singleton
            ptl = prop_type.lower()
            info_obj.set_names([ptl])

            # Handle property requests of type group
            if ptl == "group":
                info_obj.update(os.path.basename(self.Reader.filename), [[x] for x in self.Groups])

            # Handle property requests of type dataset
            elif ptl == "dataset":
                info_obj.update(
                    os.path.basename(self.Reader.filename), [[x] for x in self.Datasets])

            # Handle property requests of type attribute
            elif ptl == "attribute":
                info_obj.update(
                    os.path.basename(self.Reader.filename),
                    [["{}.{}".format(k, x)] for k, v in self.Attributes.items() for x in v])

            # Return computed information object
            return info_obj

        # Unsupported property type of this point was reached
        print("*  WARNING: argHDF5Reader does not support property with type {}.".format(prop_type))
        return info_obj



# HEADER
#               arg/DataInterface/argKeyValueFilesReader.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
# HEADER

import csv
import os
import re
import sys

from arg.Common.argInformationObject import argInformationObject
from arg.DataInterface.argDataInterfaceBase import argDataInterfaceBase


class argKeyValueFilesReader(argDataInterfaceBase):
    """A concrete data interface to a key/value file
    """

    def __init__(self, *args):
        """Constructor allows for multiple matching files in directory
        """

        # Error out on incorrect constructor signature
        if len(args) != 2:
            print("** ERROR: incorrect parameters for {} constructor. Exiting.".format(
                type(self)))
            sys.exit(1)

        # Initialize storage for all matching files
        self.Readers = []

        # Initialize storage for retrieved countents
        self.Dictionaries = []

        # Retrieve delimiter value when provided, = otherwise
        delim = args[1].get("delimiter", '=')

        # Decide whether key-value order must be reversed
        reverse = args[1].get("reverse")

        # Try to find file pattern or extension
        re_file_names = args[1].get("expression")
        if not isinstance(re_file_names, re.Pattern):
            # Look for extension when no regexp was provided
            ext = args[1].get("extension")
            if ext:
                # Build regular expression from provided file extension
                re_file_names = re.compile(r"(.*)\.{}".format(ext))

        # Search for a collection of files if regexp was found or built
        if re_file_names:
            # Assume first argument is a directory name
            dir_name = args[0]

            # Ensure that specified directory exists
            try:
                list_in_dir = os.listdir(dir_name)
            except OSError:
                print("** ERROR: specified directory {} was not found. Exiting.".format(dir_name))
                sys.exit(1)

            # Iterate over all files in specified directory
            for base_name in list_in_dir:
                # Assemble full name and ignore directories
                full_name = os.path.join(dir_name, base_name)
                if not os.path.isfile(full_name):
                    continue

                # Try to match file name with regular expression
                if re_file_names.match(base_name):
                    # File name matches regular expression
                    self.Readers.append(base_name)

                    # Append current dictionary to list of existing ones
                    self.Dictionaries.append(
                        self.parse_key_value_file(full_name, delim, reverse))

        # Otherwise assume that passed argument is file name
        else:
            # Full file name signature
            self.Readers.append(os.path.basename(args[0]))

            # Retrieve corresponding dictionary
            self.Dictionaries.append(
                self.parse_key_value_file(args[0], delim, reverse))

    @staticmethod
    def parse_key_value_file(full_name, delim='=', reverse=False):
        """Parse file with provided name and given delimiter
        """

        # Initialize storage for key<delim>values in this file
        d = {}

        # Parse file as CSV file with specified delimiter
        try:
            with open(full_name, "r") as f:
                # Instantiate CSV reader and iterate over rows
                reader = csv.reader(f,
                                    delimiter=delim,
                                    skipinitialspace=True)
                for r in reader:
                    # Ignore non-conforming rows
                    if len(r) != 2:
                        continue

                    # Store possibly reversed key:value entry
                    if reverse:
                        d[r[1].strip()] = r[0].strip()
                    else:
                        d[r[0].strip()] = r[1].strip()

        except:
            print("*  WARNING: could not file with name {}.".format(
                os.path.basename(full_name)))

        # Return retrieved directory
        return d

    def get_accessors(self):
        """Return list of base file names
        """

        return self.Readers

    def get_meta_information(self):
        """Retrieve meta-information from data
        """

        # Initialize global meta-information
        meta = []

        # Iterate over all readers and dictionaries
        for r, d in zip(self.Readers, self.Dictionaries):
            # Initialize meta-information for this reader
            r_meta = {"file name": r, "number of keys": len(d), "number of unique values": len(set(d.values()))}

            # Append local meta-information to global list
            meta.append(r_meta)

        # Return global meta-information
        return meta

    def get_property_information(self, prop_type, prop_items=None):
        """Retrieve all information about given sproperty from key/value file
           NB: Sub-property is not supported for this type of files
        """

        # Returned information is dictionary of lists of lists
        info_obj = argInformationObject("arg_dict_lists_lists")

        # Look for property item values when requested
        if prop_items:
            # Set information names
            info_obj.set_names(["key", "value"])

            # Search requested items across all accessors
            for f, d in zip(self.Readers, self.Dictionaries):
                # Update information object with property item values
                info_obj.update(
                    f, [[p, d.get(p, '')] for p in prop_items])

        else:
            # Set information names
            info_obj.set_names([prop_type])

            # Search requested items across all accessors
            for f, d in zip(self.Readers, self.Dictionaries):
                # Update information object with property values
                info_obj.update(f, [[d.get(prop_type)]])

        # Return computed information object
        return info_obj
